from typing import Callable
from gi.repository import Gtk, GLib
import pulsectl_asyncio, asyncio
from ustatus.config import ModuleConfig
from ustatus.graphics.meter import Meter
from ustatus.graphics.volume import Volume
from ustatus.module import Module
from ustatus.utils.swaymsg import get_workspaces


class VolumeModule(Module):
    def __init__(self, gtk_orientation: Gtk.Orientation, **kwargs) -> None:
        super().__init__(gtk_orientation=gtk_orientation, **kwargs)
        module_widget = VolumeModuleWidget(gtk_orientation=gtk_orientation)
        self.set_module_widget(module_widget)

        self.sinks = dict()
        self.pulse = pulsectl_asyncio.PulseAsync("ustatus")

        self.updater_task = asyncio.create_task(self._init_async())
        self.update_lock = False

    def _update_sinks(self, name_volume_dict: dict):
        # New additions and updates
        for name, volume in name_volume_dict.items():
            if name in self.sinks.keys():
                label = self.sinks[name]
                label.update(volume)
            else:
                label = SinkVolume(
                    name,
                    volume,
                    gtk_orientation=Gtk.Orientation.HORIZONTAL,
                )
                self.module_widget.add_sink_volume(label)
                self.sinks[name] = label
        # Removals
        pending_removal = []
        for name, label in self.sinks.items():
            if name not in name_volume_dict.keys():
                pending_removal.append((name, label))
        for name, label in pending_removal:
            self.module_widget.remove_sink_volume(label)
            self.sinks.pop(name)

    async def _init_async(self):
        await self.pulse.connect()
        await self._update()
        async for _ in self.pulse.subscribe_events("all"):
            await self._update()

    async def _update(self) -> None:
        if not self.update_lock:
            self.update_lock = True
            sinks = await self.pulse.sink_list()
            name_volume_dict = {
                s.name: await self.pulse.volume_get_all_chans(s) for s in sinks
            }
            self._update_sinks(name_volume_dict)
            self.queue_draw()
            self.update_lock = False


class SinkVolume(Gtk.Box):
    def __init__(
        self,
        name: str,
        volume: float,
        gtk_orientation: Gtk.Orientation,
    ) -> None:
        super().__init__()
        container = Gtk.Box(orientation=gtk_orientation, spacing=5)
        self.label = Gtk.Label(label=self._label(volume))
        self.meter = Meter(volume)
        self.meter.set_size_request(40, 1)
        self.meter.set_hexpand(True)
        self.meter.set_vexpand(False)
        self.set_tooltip_text(name)
        container.pack_start(self.meter, expand=False, fill=False, padding=0)
        container.add(self.label)
        self.set_center_widget(container)
        self.show_all()

    def update(self, volume: float) -> None:
        self.label.set_label(self._label(volume))
        self.meter.update(volume)

    def _label(self, volume: float) -> str:
        return f"{volume * 100:.0f}%"


class VolumeModuleWidget(Gtk.Box):
    def __init__(self, gtk_orientation: Gtk.Orientation):
        super().__init__(orientation=gtk_orientation, spacing=5)

    def add_sink_volume(self, sink_volume: SinkVolume):
        self.pack_start(sink_volume, expand=True, fill=True, padding=0)

    def remove_sink_volume(self, sink_volume: SinkVolume):
        self.remove(sink_volume)



from gi.repository import Gtk


class Meter(Gtk.DrawingArea):
    def __init__(self, value: float = 0) -> None:
        super().__init__()
        self.value = value
        self.connect("draw", lambda area, context: self.draw(area, context))

    def update(self, volume: float):
        self.value = volume
        self.queue_draw()

    def draw(self, area, context):
        width = area.get_allocated_width()
        height = area.get_allocated_height()

        lw_constant = 2
        lw_horizontal = lw_constant
        lw_vertical = lw_constant
        fg_color = area.get_style_context().get_color(Gtk.StateFlags.NORMAL)

        def draw_polygon(context, pts):
            x0, y0 = pts[0]
            context.move_to(x0, y0)
            for x, y in pts[1:]:
                context.line_to(x, y)
            context.close_path()

        context.set_source_rgba(
            fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha
        )
        context.set_line_width(lw_horizontal)
        context.move_to(0, 0)
        context.line_to(area.get_allocated_width(), 0)
        context.stroke()

        context.move_to(0, area.get_allocated_height())
        context.line_to(area.get_allocated_width(), area.get_allocated_height())
        context.stroke()

        context.set_line_width(lw_vertical)
        context.move_to(0, 0)
        context.line_to(0, area.get_allocated_height())
        context.stroke()

        context.set_line_width(lw_vertical)
        context.move_to(area.get_allocated_width(), 0)
        context.line_to(area.get_allocated_width(), area.get_allocated_height())
        context.stroke()

        measure = [
            (0, 0),
            (0, height),
            (min(self.value, 1) * width, height),
            (min(self.value, 1) * width, 0),
        ]
        draw_polygon(context, measure)
        context.fill()

        if self.value > 1:
            extra_measure = [
                (0, 0),
                (0, height),
                (min(self.value-1, 1) * width, height),
                (min(self.value-1, 1) * width, 0),
            ]

            context.set_source_rgba(0.8, 0, 0, 1)
            draw_polygon(context, extra_measure)
            context.fill()



import logging
from typing import Callable, Set
from dbus_next.constants import PropertyAccess
from dbus_next.message import Message
from gi.repository import Gtk, GLib, DbusmenuGtk3, Gdk
from ustatus.config import ModuleConfig
from ustatus.module import Module
from dbus_next.service import ServiceInterface, method, dbus_property, signal
from dbus_next.signature import Variant
from dbus_next.aio.message_bus import MessageBus
from multiprocessing import Process

import asyncio


class TrayModule(Module):
    def __init__(
        self, gtk_orientation: Gtk.Orientation, update_period_seconds=3, **kwargs
    ) -> None:
        self.module_widget = TrayWidget(orientation=gtk_orientation)
        super().__init__(
            module_widget=self.module_widget, gtk_orientation=gtk_orientation, **kwargs
        )

        self.init_task = asyncio.create_task(self._init_async())

        self.show_all()
        self.module_widget.show_all()

    def _update(self) -> bool:
        self.module_widget.update()
        return True

    async def _init_async(self):
        await self._init_watcher()
        await self._init_host()
        await self._attach_to_watcher()

    async def _attach_to_watcher(self):
        bus = await MessageBus().connect()
        introspection = await bus.introspect(
            "org.kde.StatusNotifierWatcher", "/StatusNotifierWatcher"
        )
        proxy_object = bus.get_proxy_object(
            "org.kde.StatusNotifierWatcher", "/StatusNotifierWatcher", introspection
        )
        interface = proxy_object.get_interface("org.kde.StatusNotifierWatcher")
        for merged_string in await interface.get_registered_status_notifier_items():
            logging.info(f"{merged_string}")
            await self._new_item(merged_string)

        interface.on_status_notifier_item_registered(
            lambda merged_string: self._new_item_callback(merged_string)
        )

        interface.on_status_notifier_item_unregistered(
            lambda merged_string: self.__item_removed__(merged_string)
        )

        await bus.wait_for_disconnect()

    def _new_item_callback(self, merged_string: str):
        asyncio.create_task(self._new_item(merged_string))

    async def _new_item(self, merged_string: str):
        bus_name, obj_path = service_path_from_merged(merged_string)
        item = await TrayItem.init(bus_name, obj_path)
        self.module_widget.new_item(item)

    def __item_removed__(self, merged_string: str):
        bus_name, obj_path = service_path_from_merged(merged_string)
        self.module_widget.remove_item(bus_name)

    async def _init_watcher(self):
        bus = await MessageBus().connect()
        interface = StatusNotifierWatcher("org.kde.StatusNotifierWatcher", bus)
        bus.export("/StatusNotifierWatcher", interface)
        asyncio.create_task(bus.request_name("org.kde.StatusNotifierWatcher"))
        logging.info("Watcher service initialized")

    async def _init_host(self):
        bus = await MessageBus().connect()
        interface = StatusNotifierHost("org.kde.StatusNotifierHost")
        bus.export("/StatusNotifierHost", interface)
        asyncio.create_task(bus.request_name("org.kde.StatusNotifierHost-ustatus"))
        logging.info("Host service initialized")


class TrayWidget(Gtk.FlowBox):
    def __init__(self, orientation=Gtk.Orientation.HORIZONTAL):
        super().__init__()
        match orientation:
            case Gtk.Orientation.HORIZONTAL:
                self.set_orientation(Gtk.Orientation.VERTICAL)
            case Gtk.Orientation.VERTICAL:
                self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_selection_mode(Gtk.SelectionMode.NONE)
        self.items = dict()

    def update(self):
        self.show_all()

    def new_item(self, item):
        self.items[item.bus_name] = item
        self.add(item)
        self.update()

    def remove_item(self, bus_name):
        if bus_name in self.items:
            item = self.items[bus_name]
            self.remove(item)


class TrayItem(Gtk.Box):
    interface_name = "org.kde.StatusNotifierItem"

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL)

    @classmethod
    async def init(cls, bus_name, obj_path):
        obj = cls()

        obj.button_icon = Gtk.Image()

        obj.button = Gtk.Button()
        obj.button.set_relief(Gtk.ReliefStyle.NONE)
        obj.button.set_image(obj.button_icon)
        obj.button.set_focus_on_click(False)

        Module.__remove_button_frame__(obj.button)
        obj.add(obj.button)
        obj.bus_name = bus_name
        obj.obj_path = obj_path
        obj.bus = await MessageBus().connect()
        introspection = await obj.bus.introspect(bus_name, obj_path)
        proxy_object = obj.bus.get_proxy_object(bus_name, obj_path, introspection)
        obj.interface = proxy_object.get_interface(cls.interface_name)
        await obj.__get_id__()
        await obj.__get_title__()
        await obj.__get_icon_name__()
        await obj.__get_menu__()

        obj.interface.on_new_title(lambda: obj.__on_new_title_callback__())
        obj.interface.on_new_icon(lambda: obj.__on_new_icon_callback__())

        return obj

    def __on_new_title_callback__(self):
        asyncio.create_task(self.__get_title__())

    def __on_new_icon_callback__(self):
        asyncio.create_task(self.__get_icon_name__())

    async def __get_id__(self):
        self.id = await self.interface.get_id()

    async def __get_title__(self):
        self.title = await self.interface.get_title()

    async def __get_icon_name__(self):
        self.icon_name = await self.interface.get_icon_name()
        self.button_icon.set_from_icon_name(self.icon_name, Gtk.IconSize.LARGE_TOOLBAR)
        self.show_all()

    async def __get_menu_path__(self):
        self.menu_path = await self.interface.get_menu()

    async def __get_menu__(self):
        await self.__get_menu_path__()
        self.menu = DbusmenuGtk3.Menu.new(self.bus_name, self.menu_path)
        self.menu.attach_to_widget(self.button)
        self.menu.set_take_focus(True)
        self.button.connect("clicked", self.__on_button_clicked__)
        self.show_all()

    def __on_button_clicked__(self, widget):
        self.menu.popup_at_widget(
            widget, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH_WEST, None
        )


class StatusNotifierWatcher(ServiceInterface):
    def __init__(self, name, bus: MessageBus):
        super().__init__(name)
        self.name = name
        self._items_registered: Set = set()
        self._hosts_registered: Set = set()
        bus.add_message_handler(self.register_sni_handler)

    def register_sni_handler(self, msg: Message):
        if msg.interface == self.name and msg.member == "RegisterStatusNotifierItem":
            self._items_registered.add(f"{msg.sender}{msg.body[0]}")
            return Message.new_method_return(msg, "", [])

    @method()
    def RegisterStatusNotifierItem(self, service: "s"):
        pass

    @method()
    def RegisterStatusNotifierHost(self, service: "s"):
        self._hosts_registered.add(service)

    @dbus_property(access=PropertyAccess.READ)
    def RegisteredStatusNotifierItems(self) -> "as":
        return list(self._items_registered)

    @dbus_property(access=PropertyAccess.READ)
    def IsStatusNotifierHostRegistered(self) -> "b":
        return len(self._hosts_registered) > 0

    @dbus_property(access=PropertyAccess.READ)
    def ProtocolVersion(self) -> "i":
        return 0


class StatusNotifierHost(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)


def service_path_from_merged(merged: str):
    split_list = merged.split("/", maxsplit=1)
    if len(split_list) == 1:
        return merged, "/StatusNotifierItem"
    elif len(split_list) == 2:
        service, path = split_list
        path = "/" + path
        return service, path
    else:
        raise Exception("Error while splitting merged service/path string")



from typing import Callable
from gi.repository import Gtk, Gdk, GLib
from ustatus.config import ModuleConfig
from ustatus.graphics.line_graph import LineGraph
from ustatus.graphics.bar_graph import BarGraph
from ustatus.module import Module
import psutil
from ustatus.cpuinfo import get_core_per_cpu
from collections import deque
import itertools


class CpuModule(Module):
    def __init__(
        self, bar_width: int, uptate_period_seconds=1, history_length=60, **kwargs
    ) -> None:
        module_widget = CpuModuleWidget(bar_width=bar_width)
        modal_widget = CpuModuleModalWidget(history_length=history_length)
        self.history_length = history_length
        self.cpu_history = History(
            updater=self.update_cpu_history, maxlen=history_length
        )
        self.freq_history = History(
            updater=self.update_freq_history, maxlen=history_length
        )
        self.temp_history = History(
            updater=self.update_temp_history, maxlen=history_length
        )
        # self.fans_history = History(
        # updater=self.update_fans_history, maxlen=history_length
        # )
        super().__init__(
            module_widget=module_widget,
            bar_width=bar_width,
            **kwargs,
        )
        modal_button = self.get_popover_menubutton(modal_widget=modal_widget)
        self.module_widget.set_popover_menubutton(modal_button)
        GLib.timeout_add(uptate_period_seconds * 1000, lambda: self._update())
        GLib.timeout_add(uptate_period_seconds * 1000, lambda: self._update_modal())

    def _update(self):
        self.cpu_history.update()
        self.freq_history.update()
        self.temp_history.update()
        # self.fans_history.update()
        self.module_widget.update(self.cpu_history.peek_one())
        return True

    def _update_modal(self):
        self.modal_widget.push_value(
            self.cpu_history.peek_one(),
            self.temp_history.peek_one(),
        )
        # self.modal_widget.push_fans_value(self.fans_history.peek_one(), "thinkpad")
        self.modal_widget.push_freq_value(self.freq_history.peek_one())
        return True

    def update_cpu_history(self):
        cpu_percents = [x / 100 for x in psutil.cpu_percent(interval=None, percpu=True)]
        return cpu_percents

    def update_freq_history(self):
        freqs = [x.current for x in psutil.cpu_freq(percpu=True)]
        return freqs

    def update_fans_history(self):
        thinkpad = psutil.sensors_fans()["thinkpad"]
        fans = []
        for fan in thinkpad:
            fans.append(fan.current)
        return fans

    def update_temp_history(self):
        coretemp = psutil.sensors_temperatures()["coretemp"]
        temps = []
        for ct in coretemp:
            temps.append(ct.current)
        return temps


class CpuModuleWidget(Gtk.Box):
    def __init__(self, bar_width: int) -> None:
        super().__init__()
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.button = None
        self.meter = BarGraph(
            n_values=psutil.cpu_count(),
            color=Gdk.RGBA(red=0.31, green=0.31, blue=0.80, alpha=1),
        )
        self.meter.set_size_request(bar_width, bar_width)
        self.add(self.meter)

    def update(self, values):
        self.meter.set_values(values)

    def set_popover_menubutton(self, button: Gtk.Widget):
        if self.button:
            self.remove(self.button)
        self.button = button
        button_image = Gtk.Image.new_from_icon_name(
            "go-down-symbolic", Gtk.IconSize.SMALL_TOOLBAR
        )
        self.button.set_image(button_image)
        Module.__remove_button_frame__(self.button)
        self.button.set_relief(Gtk.ReliefStyle.NONE)
        self.add(button)


class CpuModuleModalWidget(Gtk.Box):
    def __init__(self, history_length: int = 20) -> None:
        super().__init__()
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.history_length = history_length
        label = Gtk.Label(label="CPU Usage")
        self.add(label)
        self.init_cpu_usage()
        self.init_cpu_freq()
        self.init_coretemp()
        # self.init_fanspeed()

    def init_coretemp(self):
        temp_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        coretemp = psutil.sensors_temperatures()["coretemp"]
        self.temps = []
        for ct in coretemp:
            if ct.critical:
                crit = ct.critical
            else:
                crit = 100
            temp = Temp(label=ct.label, history_length=self.history_length, crit=crit)
            self.temps.append(temp)
            temp_container.add(temp)
        self.add(temp_container)

    def init_cpu_usage(self):
        core_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        cpu_info = get_core_per_cpu()
        core_ids = list(set(cpu_info))
        cores = [Core(id) for id in core_ids]
        for core in cores:
            core_container.add(core)
        self.cpus = [
            Cpu(i, history_length=self.history_length)
            for i in range(psutil.cpu_count())
        ]
        for i, cpu in enumerate(self.cpus):
            cores[cpu_info[i]].add_cpu(cpu)
        self.add(core_container)

    def init_cpu_freq(self):
        core_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        cpu_info = get_core_per_cpu()
        core_ids = list(set(cpu_info))
        cores = [Core(id) for id in core_ids]
        for core in cores:
            core_container.add(core)
        self.freqs = [
            Freq(i, history_length=self.history_length, max_freq=freq.max)
            for i, freq in enumerate(psutil.cpu_freq(percpu=True))
        ]
        for i, cpu in enumerate(self.freqs):
            cores[cpu_info[i]].add_cpu(cpu)
        self.add(core_container)

    def init_fanspeed(self):
        self.fans = dict()
        fanspeed_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        fan_dict = psutil.sensors_fans()
        for key, val in fan_dict.items():
            self.fans[key] = []
            for i, fan in enumerate(val):
                if fan.label:
                    label = fan.label
                else:
                    label = f"{key} {i}"
                fan_gtk = FanSpeed(label=label, history_length=self.history_length)
                self.fans[key].append(fan_gtk)
                fanspeed_container.add(fan_gtk)

        self.add(child=fanspeed_container)

    def set_values(self, cpu_values, temp_values):
        for v, cpu in zip(cpu_values, self.cpus):
            cpu.set_values(v)
        for v, temp in zip(temp_values, self.temps):
            temp.set_values(v)

    def set_freq_values(self, freq_values):
        for v, freq in zip(freq_values, self.freqs):
            freq.set_values(v)

    def push_freq_value(self, freq_values):
        for v, freq in zip(freq_values, self.freqs):
            freq.push_value(v)

    def set_fans_values(self, fan_values, fan_key):
        for v, fan in zip(fan_values, self.fans[fan_key]):
            fan.set_values(v)

    def push_fans_value(self, fan_values, fan_key):
        for v, fan in zip(fan_values, self.fans[fan_key]):
            fan.push_value(v)

    def push_value(self, cpu_value, temp_value):
        for v, cpu in zip(cpu_value, self.cpus):
            cpu.push_value(v)
        for v, temp in zip(temp_value, self.temps):
            temp.push_value(v)


class History:
    def __init__(
        self,
        updater: Callable,
        maxlen: int = 100,
    ):
        self.maxlen = maxlen
        seed = [0] * maxlen
        self.history = [deque(seed, maxlen=maxlen) for _ in range(psutil.cpu_count())]
        self.updater = updater

    def push_value(self, new_value):
        for value, thread_history in zip(new_value, self.history):
            thread_history.popleft()
            thread_history.append(value)

    def peek_one(self):
        return [thread_history[-1] for thread_history in self.history]

    def peek_n(self, n):
        return [
            list(itertools.islice(thread_history, self.maxlen - n, self.maxlen))
            for thread_history in self.history
        ]

    def update(self):
        new_value = self.updater()
        self.push_value(new_value)


class Cpu(Gtk.Box):
    def __init__(self, cpu_id: int, history_length: int = 20) -> None:
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.label = Gtk.Label(label=self.__label__(cpu_id))
        self.curr_label = Gtk.Label(label="")
        self.meter = LineGraph(
            n_values=history_length,
            color=Gdk.RGBA(red=0.31, green=0.31, blue=0.80, alpha=1),
        )
        self.meter.set_size_request(100, 100)
        self.add(self.meter)
        self.add(self.curr_label)
        self.add(self.label)

    def set_values(self, new_values: list[float]) -> None:
        self.meter.set_values(new_values)
        self.curr_label.set_label(f"{new_values[-1]*100:.1f}%")

    def push_value(self, new_value: float) -> None:
        self.meter.push_value(new_value)
        self.curr_label.set_label(f"{new_value * 100:.1f}%")

    def __label__(self, cpu_id: int) -> str:
        return f"CPU {cpu_id}"


class Freq(Gtk.Box):
    def __init__(self, cpu_id: int, history_length: int = 20, max_freq=3000) -> None:
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.label = Gtk.Label(label=self.__label__(cpu_id))
        self.curr_label = Gtk.Label(label="")
        self.meter = LineGraph(
            n_values=history_length,
            color=Gdk.RGBA(red=0.21, green=0.60, blue=0.60, alpha=1),
        )
        self.meter.set_size_request(100, 100)
        self.set_max_freq(max_freq)
        self.add(self.meter)
        self.add(self.curr_label)
        self.add(self.label)

    def set_max_freq(self, max_freq):
        self.max_freq = max_freq
        self.meter.set_max(self.max_freq)

    def set_values(self, new_values: list[float]) -> None:
        self.meter.set_values(new_values)
        self.curr_label.set_label(f"{new_values[-1]:.1f} MHz")

    def push_value(self, new_value: float) -> None:
        self.meter.push_value(new_value)
        self.curr_label.set_label(f"{new_value:.1f} MHz")

    def __label__(self, cpu_id: int) -> str:
        return f"CPU {cpu_id}"


class Temp(Gtk.Box):
    def __init__(self, label: str, history_length: int = 20, crit: float = 100) -> None:
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.label = Gtk.Label(label=label)
        self.curr_label = Gtk.Label(label="")
        self.crit = crit
        self.meter = LineGraph(
            n_values=history_length,
            color=Gdk.RGBA(red=0.80, green=0.31, blue=0.31, alpha=1),
        )
        self.meter.set_size_request(100, 100)
        self.meter.set_max(crit)
        self.add(self.meter)
        self.add(self.curr_label)
        self.add(self.label)

    def set_values(self, new_values: list[float]) -> None:
        self.meter.set_values(new_values)
        self.curr_label.set_label(f"{new_values[-1]:.1f}°C")

    def push_value(self, new_value: float) -> None:
        self.meter.push_value(new_value)
        self.curr_label.set_label(f"{new_value:.1f}°C")


class FanSpeed(Gtk.Box):
    def __init__(self, label: str, history_length: int = 20) -> None:
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.history_length = history_length
        self.label = Gtk.Label(label=label)
        self.label.connect(
            "size-allocate",
            lambda label, size: label.set_size_request(size.width - 1, -1),
        )
        self.curr_label = Gtk.Label(label="")
        self.meter = LineGraph(
            n_values=history_length,
            color=Gdk.RGBA(red=0.31, green=0.80, blue=0.31, alpha=1),
        )
        self.meter.set_size_request(100, 100)
        self.add(self.meter)
        self.add(self.curr_label)
        self.add(self.label)
        self.running_max = 1
        self.running_max_age = 0

    def set_values(self, new_values: list[float]) -> None:
        assert len(new_values) == self.history_length
        self.running_max = max(new_values)
        self.running_max_age = 0
        self.meter.set_values(new_values)
        self.meter.set_max(self.running_max)
        self.curr_label.set_label(f"{new_values[-1]:d} RPM")

    def push_value(self, new_value: float) -> None:
        self.meter.push_value(new_value)
        self.curr_label.set_label(f"{new_value:d} RPM")

        self.running_max_age += 1
        if new_value > self.running_max:
            self.running_max = new_value
            self.running_max_age = 0
            self.meter.set_max(self.running_max)
        elif self.running_max_age > self.history_length:
            self.running_max = max(self.meter.values)
            self.running_max_age = 0
            self.meter.set_max(self.running_max)


class Core(Gtk.Box):
    def __init__(self, core_id: int) -> None:
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.core_id = core_id
        self.cpu_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5)
        self.label = Gtk.Label(label=self.__label__(core_id))
        self.add(self.cpu_container)
        self.add(self.label)

    def add_cpu(self, cpu_widget: Gtk.Widget) -> None:
        self.cpu_container.add(cpu_widget)

    def __label__(self, core_id: int) -> str:
        return f"Core {core_id}"



import logging
from gi.repository import Gtk
from ustatus.module import Module
from dbus_next.aio.message_bus import MessageBus
from dbus_next.constants import BusType
from dbus_next.errors import DBusError
from ustatus.utils.notifications import notify_error

import asyncio


class PowerProfilesModule(Module):
    def __init__(
        self,
        **kwargs,
    ) -> None:

        self.active_profile = "unknown"

        self.modal_widget = PowerProfilesModalWidget(
            set_profile_callback=lambda profile: asyncio.create_task(
                self._set_active_profile(profile)
            )
        )
        modal_menubutton = self.get_popover_menubutton(self.modal_widget)
        self.module_widget = PowerProfilesWidget(modal_menubutton, self.active_profile)

        super().__init__(module_widget=self.module_widget, **kwargs)

        self.init_dbus_task = asyncio.create_task(self._init_dbus())

    def _update(self) -> bool:
        return True

    def _update_modal(self) -> bool:
        return True

    async def _sync_profiles(self):
        self.profiles = await self.dbus_interface.get_profiles()
        self.active_profile = await self.dbus_interface.get_active_profile()

        logging.info(f"Active profile: {self.active_profile}")

        self.module_widget.set_active_profile(self.active_profile)
        self.modal_widget.set_profiles(self.profiles, self.active_profile)

    async def _set_active_profile(self, profile):
        try:
            await self.dbus_interface.set_active_profile(profile)
        except DBusError as e:
            notify_error(body=f"Error trying to change active profile: {e}")

    async def _init_dbus(self):
        bus_name = "net.hadess.PowerProfiles"
        obj_path = "/net/hadess/PowerProfiles"
        interface_name = "net.hadess.PowerProfiles"

        self.bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
        introspection = await self.bus.introspect(bus_name, obj_path)
        proxy_object = self.bus.get_proxy_object(bus_name, obj_path, introspection)
        self.dbus_interface = proxy_object.get_interface(interface_name)

        self.properties_interface = proxy_object.get_interface(
            "org.freedesktop.DBus.Properties"
        )
        self.properties_interface.on_properties_changed(
            lambda _a, _b, _c: asyncio.create_task(self._sync_profiles())
        )

        await self._sync_profiles()


class PowerProfilesWidget(Gtk.Grid):
    def __init__(self, modal_menubutton, active_profile: str):
        super().__init__()

        self.set_hexpand(True)
        self.set_vexpand(True)

        self.active_profile_label = Gtk.Label()
        self.active_profile_label.set_hexpand(True)
        self.set_active_profile(active_profile)

        self.menubutton = modal_menubutton
        menubutton_image = Gtk.Image.new_from_icon_name(
            "go-down-symbolic", Gtk.IconSize.SMALL_TOOLBAR
        )
        self.menubutton.set_image(menubutton_image)
        Module.__remove_button_frame__(self.menubutton)
        self.menubutton.set_relief(Gtk.ReliefStyle.NONE)

        self.attach(self.active_profile_label, 0, 0, 1, 1)
        self.attach(self.menubutton, 0, 1, 1, 1)

    def set_active_profile(self, active_profile: str):
        self.active_profile = active_profile
        self.active_profile_label.set_label(active_profile)


class PowerProfilesModalWidget(Gtk.Box):
    def __init__(self, set_profile_callback):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.set_profile_callback = set_profile_callback
        self.show_all()

    def _set_active_profile(self, button, profile):
        if button.get_active():
            logging.info(f"Setting active profile {profile}")
            self.set_profile_callback(profile)

    def set_profiles(self, profiles, active_profile):
        self.foreach(lambda child: self.remove(child))
        profile_button = None
        for profile in profiles:
            logging.info(f"Creating button for profile {profile['Profile'].value}")
            profile_button = self._create_radiobutton(
                profile_button, profile, active_profile
            )
            self.add(profile_button)
        self.show_all()

    def _create_radiobutton(self, prev_button, profile, active_profile):
        button = Gtk.RadioButton.new_with_label_from_widget(
            prev_button, profile["Profile"].value
        )
        if profile["Profile"].value == active_profile:
            button.set_active(True)
        else:
            button.set_active(False)
        button.connect("toggled", self._set_active_profile, profile["Profile"].value)
        return button



def get_core_per_cpu():
    cpuinfo = open("/proc/cpuinfo", "r").read().strip().split("\n\n")
    cores = []
    for cpu in cpuinfo:
        [core] = [
            int(s.split(":")[1].strip())
            for s in cpu.splitlines()
            if s.startswith("core id")
        ]
        cores.append(core)
    return cores



from typing import Callable, Optional
from gi.repository import Gtk, GLib

from ustatus.config import BarConfig, ModuleConfig


class Module(Gtk.Box):
    def __init__(
        self,
        gtk_orientation: Gtk.Orientation,
        toggle_modal: Callable,
        module_config: ModuleConfig,
        bar_config: BarConfig,
        bar_width: int,
        output,
        module_widget: Optional[Gtk.Widget] = None,
    ) -> None:
        super().__init__(orientation=gtk_orientation)
        self.gtk_orientation = gtk_orientation
        self.toggle_modal = toggle_modal
        self.bar_width = bar_width
        self.config = module_config
        self.output = output
        if self.config.show_label:
            self.add(Gtk.Label(label=self.config.label))
        if module_widget:
            self.set_module_widget(module_widget)
        self._setup_css()

    def _update(self):
        raise NotImplementedError()

    def set_module_widget(self, module_widget):
        self.module_widget = module_widget
        self.module_widget.get_style_context().add_class("module-widget")
        self.add(self.module_widget)

    def _expand_widthwise(self, widget=None):
        if widget is None:
            widget = self
        match self.gtk_orientation:
            case Gtk.Orientation.HORIZONTAL:
                widget.set_vexpand(True)
            case Gtk.Orientation.VERTICAL:
                widget.set_hexpand(True)

    def get_popover_menubutton(self, modal_widget: Gtk.Widget):
        self.modal_widget = modal_widget
        button = Gtk.Button()
        button.connect("clicked", lambda _: self.toggle_modal(self.modal_widget))
        modal_widget.show_all()

        return button

    def _setup_css(self):
        style_context = self.get_style_context()
        style_context.add_class("module")
        style_context.add_class(self.config.type)

    @staticmethod
    def __remove_button_frame__(button):
        button_style_context = button.get_style_context()
        button_style_context.add_class("module-button")


class ModuleWithModal(Module):
    def __init__(
        self,
        module_widget: Gtk.Widget,
        modal_widget: Gtk.Widget,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        button = self.get_popover_menubutton(modal_widget)
        Module.__remove_button_frame__(button)
        button.add(module_widget)
        self.set_module_widget(button)
        self.module_widget = module_widget

    def _update_modal(self):
        raise NotImplementedError()



import asyncio
import logging
import os
import gi
from python_reactive_ui import Component
from python_reactive_ui.backends.gtk3.root import create_root
from ustatus.components.module import ReactiveModule
from ustatus.components.modules.battery import Battery
from ustatus.components.modules.clock import Clock
from ustatus.components.modules.test import Test
from ustatus.components.modules.volume import Volume

from ustatus.config import BarConfig, Config, ConfigError
from ustatus.module import Module
from ustatus.modules.cpu_module import CpuModule
from ustatus.modules.mpris_module import MprisModule
from ustatus.modules.power_module import PowerModule
from ustatus.modules.power_profiles_module import PowerProfilesModule
from ustatus.modules.sway_module import SwayModule
from ustatus.modules.tray_module import TrayModule
from ustatus.modules.volume_module import VolumeModule
from ustatus.remote_service import init_service
from ustatus.utils.swaymsg import get_outputs

gi.require_version("Gtk", "3.0")
gi.require_version("DbusmenuGtk3", "0.4")
gi.require_version("GtkLayerShell", "0.1")

from gi.repository import Gtk, GtkLayerShell, Gdk, Notify


class Ustatus(Gtk.Application):
    def do_activate(self):
        self.config: Config = Config()
        bar_name = self.config.bar_name
        output = self.config.config_dict["bars"][bar_name].get("output", None)
        Notify.init(f"ustatus {bar_name}")

        # Create the windows and link them to current application
        self._create_windows()

        self.bar_config: BarConfig = self.config.get_bar_config(bar_name)
        self.modal_widget = None
        self.bar_name = bar_name
        self.output = output

        self._setup_gtk_theme()
        self._setup_css_classes()
        self._load_css()

        self._init_box()
        self._init_layer_shell()
        self._init_modules()

        self.center_box.show_all()
        self.box.show_all()
        self.window.show_all()

        # self.connect("destroy", Gtk.main_quit)

        asyncio.create_task(
            init_service(
                lambda: self.hide_status(), lambda: self.show_status(), bar_name
            )
        )

        if self.output:
            asyncio.create_task(self._move_to_monitor())

        logging.info(f"Initialized bar {bar_name}")

    def _create_windows(self):
        self.window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
        self.window.set_application(self)
        self.modal_window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
        self.modal_window.set_application(self)

    def _setup_gtk_theme(self):
        if self.bar_config.theme_override is not None:
            logging.info(
                f"Overriding default theme with {self.bar_config.theme_override}"
            )
            Gtk.Settings.get_default().set_property(
                "gtk_theme_name", self.bar_config.theme_override
            )

    def _setup_css_classes(self):
        style_context = self.window.get_style_context()
        style_context.add_class(self.bar_name)
        style_context.add_class("ustatus")

        modal_style_context = self.modal_window.get_style_context()
        modal_style_context.add_class(self.bar_name)
        modal_style_context.add_class("ustatus")
        modal_style_context.add_class("modal")

    async def _move_to_monitor(self):
        monitor = await self._get_gdk_monitor(self.output)
        GtkLayerShell.set_monitor(self, monitor)

    async def _get_gdk_monitor(self, output):
        outputs = await get_outputs()
        out = list(filter(lambda m: m["name"] == output, outputs))
        assert len(out) == 1
        out = out[0]

        display = self.get_display()
        for num in range(display.get_n_monitors()):
            m = display.get_monitor(num)
            if m.get_model() == out["model"]:
                return m

        raise Exception(f"Could not find monitor {out}")

    def _init_layer_shell(self):
        GtkLayerShell.init_for_window(self.window)
        for anchor in self.bar_config.anchors:
            match anchor:
                case "right":
                    GtkLayerShell.set_anchor(self.window, GtkLayerShell.Edge.RIGHT, 1)
                case "left":
                    GtkLayerShell.set_anchor(self.window, GtkLayerShell.Edge.LEFT, 1)
                case "top":
                    GtkLayerShell.set_anchor(self.window, GtkLayerShell.Edge.TOP, 1)
                case "bottom":
                    GtkLayerShell.set_anchor(self.window, GtkLayerShell.Edge.BOTTOM, 1)
                case _:
                    raise ConfigError(f"Anchor point {anchor} not defined.")
        GtkLayerShell.init_for_window(self.modal_window)
        self._update_modal_anchor()
        self.window.connect(
            "size-allocate", lambda window, size: self._update_modal_anchor()
        )
        if self.bar_config.exclusive:
            self._update_exclusive_zone()
            self.window.connect(
                "size-allocate", lambda window, size: self._update_exclusive_zone()
            )

    def _init_modules(self):
        self.modules_start = self.instantiate_modules(self.bar_config.modules_start)
        self.modules_center = self.instantiate_modules(self.bar_config.modules_center)
        self.modules_end = self.instantiate_modules(
            reversed(self.bar_config.modules_end)
        )

        for module in self.modules_start:
            self.box.pack_start(child=module, expand=False, fill=False, padding=0)
        for module in self.modules_center:
            self.center_box.pack_start(
                child=module, expand=False, fill=False, padding=0
            )
        for module in self.modules_end:
            self.box.pack_end(child=module, expand=False, fill=False, padding=0)

    def show_status(self):
        self.window.show()

    def hide_status(self):
        self.hide_modal()
        self.window.hide()

    def show_modal(self, widget: Gtk.Widget):
        if self.modal_widget:
            self.modal_window.remove(self.modal_widget)
        self.modal_widget = widget
        self.modal_widget.get_style_context().add_class("modal-widget")
        self.modal_window.add(self.modal_widget)
        self.modal_window.show_all()

    def hide_modal(self):
        if self.modal_widget:
            self.modal_window.remove(self.modal_widget)
            self.modal_widget = None
        self.modal_window.hide()

    def toggle_modal(self, widget: Gtk.Widget):
        if self.modal_widget == widget:
            self.hide_modal()
        else:
            self.show_modal(widget)

    def _init_box(self):
        self.scrolled_window_container = Gtk.ScrolledWindow.new()
        self.window.add(self.scrolled_window_container)
        self.box = Gtk.Box()
        self.box.set_vexpand(True)
        self.center_box = Gtk.Box()
        self.box.set_center_widget(self.center_box)
        if self.bar_config.width == "auto":
            h_scroll_policy = Gtk.PolicyType.NEVER
        else:
            h_scroll_policy = Gtk.PolicyType.EXTERNAL
            self.scrolled_window_container.set_max_content_width(self.bar_config.width)
            self.scrolled_window_container.set_min_content_width(self.bar_config.width)

        if self.bar_config.height == "auto":
            v_scroll_policy = Gtk.PolicyType.NEVER
        else:
            v_scroll_policy = Gtk.PolicyType.EXTERNAL
            self.scrolled_window_container.set_max_content_height(
                self.bar_config.height
            )
            self.scrolled_window_container.set_min_content_height(
                self.bar_config.height
            )

        self.scrolled_window_container.set_policy(h_scroll_policy, v_scroll_policy)

        match self.bar_config.orientation:
            case "horizontal":
                self.gtk_orientation = Gtk.Orientation.HORIZONTAL
                self.box.set_orientation(self.gtk_orientation)
                self.center_box.set_orientation(self.gtk_orientation)
            case "vertical":
                self.gtk_orientation = Gtk.Orientation.VERTICAL
                self.box.set_orientation(self.gtk_orientation)
                self.center_box.set_orientation(self.gtk_orientation)
            case other:
                raise ConfigError(f"Orientation {other} not defined.")
        self.scrolled_window_container.add(self.box)

    def instantiate_modules(self, module_names):
        modules = []
        for module_name in module_names:
            module_config = self.config.get_module_config(module_name)
            modules.append(
                self._create_builtin_module(
                    module_config=module_config,
                    bar_config=self.bar_config,
                    gtk_orientation=self.gtk_orientation,
                    toggle_modal=self.toggle_modal,
                    output=self.output,
                    bar_width=self.bar_config.width,
                )
            )
            if self.bar_config.separators:
                modules.append(Gtk.Separator.new(orientation=self._not_orientation()))
        return modules

    def _update_modal_anchor(self):
        margin_x = 1
        margin_y = 1
        match self.gtk_orientation:
            case Gtk.Orientation.VERTICAL:
                margin_x = self.window.get_allocated_width()
                margin_y = 1
            case Gtk.Orientation.HORIZONTAL:
                margin_y = self.window.get_allocated_height()
                margin_x = 1
        for anchor in self.bar_config.anchors:
            match anchor:
                case "right":
                    GtkLayerShell.set_anchor(
                        self.modal_window, GtkLayerShell.Edge.RIGHT, True
                    )
                    GtkLayerShell.set_margin(
                        self.modal_window, GtkLayerShell.Edge.RIGHT, margin_x
                    )
                case "left":
                    GtkLayerShell.set_anchor(
                        self.modal_window, GtkLayerShell.Edge.LEFT, True
                    )
                    GtkLayerShell.set_margin(
                        self.modal_window, GtkLayerShell.Edge.LEFT, margin_x
                    )
                case "top":
                    GtkLayerShell.set_anchor(
                        self.modal_window, GtkLayerShell.Edge.TOP, True
                    )
                    GtkLayerShell.set_margin(
                        self.modal_window, GtkLayerShell.Edge.TOP, margin_y
                    )
                case "bottom":
                    GtkLayerShell.set_anchor(
                        self.modal_window, GtkLayerShell.Edge.BOTTOM, True
                    )
                    GtkLayerShell.set_margin(
                        self.modal_window, GtkLayerShell.Edge.BOTTOM, margin_y
                    )
                case _:
                    raise ConfigError(f"Anchor point {anchor} not defined.")

    def _update_exclusive_zone(self):
        match self.gtk_orientation:
            case Gtk.Orientation.VERTICAL:
                margin = self.window.get_allocated_width()
            case Gtk.Orientation.HORIZONTAL:
                margin = self.window.get_allocated_height()
            case other:
                raise Exception(f"Orientation {other} not recognized")
        GtkLayerShell.set_exclusive_zone(self.window, margin)

    def _not_orientation(self):
        match self.gtk_orientation:
            case Gtk.Orientation.VERTICAL:
                return Gtk.Orientation.HORIZONTAL
            case Gtk.Orientation.HORIZONTAL:
                return Gtk.Orientation.VERTICAL
            case other:
                raise Exception(f"Orientation {other} not recognized")

    def _load_css(self):
        self._load_custom_css()
        screen = Gdk.Screen.get_default()
        provider = Gtk.CssProvider()
        style_context = Gtk.StyleContext()
        style_context.add_provider_for_screen(
            screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )
        css = """
    .module-button {padding: 0; border: none;}
    """
        provider.load_from_data(css.encode())

    def _load_custom_css(self):
        if "XDG_CONFIG_HOME" in os.environ:
            config_path = os.path.expandvars("$XDG_CONFIG_HOME/ustatus/ustatus.css")
        else:
            config_path = os.path.expandvars("$HOME/.config/ustatus/ustatus.css")
        if os.path.exists(config_path):
            logging.info(f"Loading CSS from {config_path}")
            screen = Gdk.Screen.get_default()
            provider = Gtk.CssProvider()
            style_context = Gtk.StyleContext()
            style_context.add_provider_for_screen(
                screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_USER
            )
            provider.load_from_path(config_path)

    def _create_builtin_module(self, **kwargs):
        builtins = {
            "volume": Volume,
            "battery": Battery,
            "mpris": MprisModule,
            "cpu": CpuModule,
            "tray": TrayModule,
            "sway": SwayModule,
            "power_profiles": PowerProfilesModule,
            "power": PowerModule,
            "test": Test,
            "clock": Clock,
        }
        module_config = kwargs["module_config"]
        if module_config.type in builtins:
            module_type = builtins[module_config.type]
            if issubclass(module_type, Module):
                return builtins[module_config.type](**kwargs)
            elif issubclass(module_type, Component):
                box = Gtk.Box()
                root = create_root(box)
                root.render(ReactiveModule(kwargs, [module_type(kwargs)]))
                return box
        else:
            raise ConfigError(f"Module type {module_config.type} not defined.")



import asyncio
from typing import Callable

from python_reactive_ui import Component
from python_reactive_ui.lib.hooks import use_effect


def use_timer(
    component: Component, timer_secs: float, callback: Callable, repeat: bool = True
):
    def set_timer():
        # get loop as a kwarg or take the default one
        loop = asyncio.get_event_loop()
        # record the loop's time when call_periodic was called
        start = loop.time()

        def run(handle):
            # XXX: we could record before = loop.time() and warn when callback(*args) took longer than interval
            # call callback now (possibly blocks run)
            callback()
            # reschedule run at the soonest time n * interval from start
            # re-assign delegate to the new handle
            if repeat:
                handle.delegate = loop.call_later(
                    timer_secs - ((loop.time() - start) % timer_secs), run, handle
                )

        class PeriodicHandle:  # not extending Handle, needs a lot of arguments that make no sense here
            def __init__(self):
                self.delegate = None

            def cancel(self):
                assert isinstance(
                    self.delegate, asyncio.Handle
                ), "no delegate handle to cancel"
                self.delegate.cancel()

        periodic = (
            PeriodicHandle()
        )  # can't pass result of loop.call_at here, it needs periodic as an arg to run
        # set the delegate to be the Handle for call_at, causes periodic.cancel() to cancel the call to run
        periodic.delegate = loop.call_at(start + timer_secs, run, periodic)
        # return the 'wrapperk'
        return periodic

    use_effect(component, set_timer, [])



from argparse import ArgumentError
from typing import Dict, List, Optional, Type


def primitive(
    type: str = "string", description: Optional[str] = None, default=None
) -> Dict[str, str]:
    obj = dict({"type": type})
    if description is not None:
        obj["description"] = description
    if default is not None:
        obj["default"] = default
    return obj


def cmdline_friendly(type: Dict) -> bool:
    if "type" in type:
        match type["type"]:
            case "string":
                return True
            case "boolean":
                return True
            case "integer":
                return True
            case "integer":
                return True
            case "array":
                match type["items"]["type"]:
                    case "string":
                        return True
    return False


def get_python_type(type: Dict) -> Type:
    if "type" in type:
        match type["type"]:
            case "string":
                return str
            case "boolean":
                return bool
            case "integer":
                return int
            case "integer":
                return int
            case "array":
                match type["items"]["type"]:
                    case "string":
                        return lambda string: string.split(",")


def string(description: Optional[str] = None, default=None) -> Dict:
    return primitive("string", description, default)


def string_enum(values: List[str], description: Optional[str] = None, default=None):
    obj = string(description=description, default=default)
    obj["enum"] = values
    return obj


def any_of(types: List, description: Optional[str] = None, default=None) -> Dict:
    obj: Dict = dict()
    obj["anyOf"] = types
    if description is not None:
        obj["description"] = description
    if default is not None:
        obj["default"] = default
    return obj


def boolean(description: Optional[str] = None, default=None) -> Dict[str, str]:
    return primitive("boolean", description, default)


def integer(description: Optional[str] = None, default=None) -> Dict[str, str]:
    return primitive("integer", description, default)


def array(items=string(), description: Optional[str] = None, default=None):
    obj = dict({"type": "array", "items": items})
    if description is not None:
        obj["description"] = description
    if default is not None:
        obj["default"] = default
    return obj


bar_size = any_of([integer(), string_enum(["auto"])], default="auto")

bar = {
    "type": "object",
    "properties": {
        "anchors": array(string()),
        "output": string(
            description="Force bar to appear in a given output, or leave empty for auto-choose"
        ),
        "orientation": string(),
        "modules_start": array(string(), default=[]),
        "modules_center": array(string(), default=[]),
        "modules_end": array(string(), default=[]),
        "exclusive": boolean(default=False),
        "width": bar_size,
        "height": bar_size,
        "separators": boolean(default=False),
        "theme_override": string(),
    },
    "required": ["anchors", "orientation"],
}

module = {
    "type": "object",
    "properties": {
        "type": string(),
        "show_label": boolean(default=False),
        "label": string(default="Label"),
        "length": integer(default=25),
    },
    "required": ["type"],
}

schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "https://example.com/product.schema.json",
    "title": "Ustatus config",
    "description": "Configuration",
    "type": "object",
    "properties": {"bars": {"type": "object", "additionalProperties": bar}},
    "properties": {"modules": {"type": "object", "additionalProperties": module}},
}



import argparse
import logging
from typing import Dict
from jsonschema import validate
import tomli
import os.path
import os
import copy
from ustatus.schema import cmdline_friendly, schema, bar, get_python_type, module


class Config:
    def __init__(self):
        self.config_dict = dict({"bars": dict(), "modules": dict()})
        self._update_with_config_files()
        self._update_with_arguments()
        self._update_with_defaults()
        validate(instance=self.config_dict, schema=schema)

    def _inherit_close(self, curr_dict, toml_dict):
        while "inherit" in curr_dict:
            inherited = copy.deepcopy(toml_dict[curr_dict["inherit"]])
            curr_dict.pop("inherit")
            curr_dict = merge_configs(source=curr_dict, destination=inherited)
        return curr_dict

    def _update_with_config_files(self):
        paths = ["examples/ustatus.toml", get_user_config_path()]
        while paths:
            new_config = read_config_from_path(paths.pop())
            if new_config:
                self.config_dict = merge_configs(new_config, self.config_dict)

    def _update_with_arguments(self):
        parser = argparse.ArgumentParser(
            description="Start a ustatus instance of the bar of given name."
        )
        parser.add_argument(
            "bar_name",
            metavar="<bar>",
            type=str,
            help="name of the bar to spawn",
        )
        for prop_name, prop_details in bar["properties"].items():
            if cmdline_friendly(prop_details):
                parser.add_argument(
                    f"--{prop_name}",
                    metavar=f"<{prop_name}>",
                    type=get_python_type(prop_details),
                    default=None,
                    help=f"({prop_details['type']}) {prop_details.get('description', '')}",
                )
        args = parser.parse_args()
        self.bar_name = args.bar_name
        if self.bar_name not in self.config_dict["bars"]:
            self.config_dict["bars"][self.bar_name] = dict()
        for prop_name, prop_details in bar["properties"].items():
            if cmdline_friendly(prop_details):
                arg = args.__getattribute__(prop_name)
                if arg is not None:
                    self.config_dict["bars"][self.bar_name][prop_name] = arg

    def _update_with_defaults(self):
        for bar_config in self.config_dict["bars"].values():
            for prop_name, prop_details in bar["properties"].items():
                if prop_name not in bar_config:
                    bar_config[prop_name] = prop_details.get("default", None)

        for module_config in self.config_dict["modules"].values():
            for prop_name, prop_details in module["properties"].items():
                if prop_name not in module_config:
                    module_config[prop_name] = prop_details.get("default", None)

    def get_bar_config(self, bar_name):
        return BarConfig(bar_name, self.config_dict)

    def get_module_config(self, module_name):
        return ModuleConfig(module_name, self.config_dict)


class ModuleConfig:
    def __init__(self, module_name: str, config_dict: Dict):
        self.module_name = module_name
        self.config_dict = config_dict

    def __getattr__(self, name):
        return self.config_dict["modules"][self.module_name][name]


class BarConfig:
    def __init__(self, bar_name: str, config_dict: Dict):
        self.bar_name = bar_name
        self.config_dict = config_dict

    def __getattr__(self, name):
        return self.config_dict["bars"][self.bar_name][name]


class ConfigError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"ConfigError: {self.message}"


def merge_configs(source, destination):
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            merge_configs(value, node)
        else:
            destination[key] = value

    return destination


def read_config_from_path(path: str):
    if os.path.exists(path):
        try:
            file = open(path, "rb")
        except:
            logging.error(f"Failed to open file {path}")
            return None
        try:
            toml_dict = tomli.load(file)
        except:
            logging.error(f"Failed to parse TOML file {path}. Exiting...")
            exit(1)
        file.close()
        logging.info(f"Loaded config file {path}")
        return toml_dict
    else:
        return None


def get_user_config_path():
    if "XDG_CONFIG_HOME" in os.environ:
        return os.path.expandvars("$XDG_CONFIG_HOME/ustatus/ustatus.toml")
    else:
        return os.path.expandvars("$HOME/.config/ustatus/ustatus.toml")



from typing import Optional
from gi.repository import Gtk, Gdk


class BarGraph(Gtk.DrawingArea):
    def __init__(self, n_values=4, color: Optional[Gdk.RGBA] = None) -> None:
        super().__init__()
        self.n_values = n_values
        self.values = [0] * n_values
        self.color = color
        self.connect("draw", lambda area, context: self.draw_bar_graph(area, context))

    def set_values(self, new_values):
        self.values = new_values
        self.queue_draw()

    def draw_bar_graph(self, area, context):
        context.scale(area.get_allocated_width(), area.get_allocated_height())
        if not self.color:
            fg_color = area.get_style_context().get_color(Gtk.StateFlags.NORMAL)
        else:
            fg_color = self.color
        context.set_line_width(0.03)

        separation: float = 1 / (self.n_values)

        for i in range(self.n_values):
            x = i * separation
            height = self.values[i]

            context.rectangle(x, 1 - height, separation, height)
            context.set_source_rgba(
                fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha
            )
            context.stroke_preserve()
            context.set_source_rgba(
                fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha / 2
            )
            context.fill()



import logging
from typing import Any, Callable, Dict, Iterable, Optional, Set
from dbus_next.aio.message_bus import MessageBus
from dbus_next.constants import BusType
from gi.repository import Gtk
from ustatus.module import Module

import asyncio


class MprisModule(Module):
    def __init__(
        self,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self._expand_widthwise()

        self.modal_widget = MprisModalWidget(self._on_select_player_callback)
        modal_menubutton = self.get_popover_menubutton(self.modal_widget)
        self.module_widget = MprisWidget(
            modal_menubutton, expander=lambda widget: self._expand_widthwise(widget)
        )
        self._expand_widthwise(self.module_widget)

        self.bus_names: Set[str] = set()
        self.init_dbus_task = asyncio.create_task(self.__init_dbus__())

        self.set_module_widget(self.module_widget)

    def _update(self) -> bool:
        return True

    def _update_modal(self) -> bool:
        return True

    def _on_name_owner_changed_callback(
        self, name: str, old_owner: str, new_owner: str
    ):
        if name.startswith("org.mpris.MediaPlayer2"):
            if new_owner:
                self.bus_names.add(name)
            else:
                self.bus_names.remove(name)
                if self.selected_player == name:
                    self._on_select_player_callback(None)
            self.modal_widget.set_items(self.bus_names)

    def _on_play_pause_callback(self):
        if self.mpris_interface:
            asyncio.create_task(self._on_play_pause())

    async def _on_play_pause(self):
        await self.mpris_interface.call_play_pause()

    def _on_next_callback(self):
        if self.mpris_interface:
            asyncio.create_task(self._on_next())

    async def _on_next(self):
        await self.mpris_interface.call_next()

    def _on_prev_callback(self):
        if self.mpris_interface:
            asyncio.create_task(self._on_prev())

    async def _on_prev(self):
        await self.mpris_interface.call_previous()

    def _on_properties_changed_callback(self, bus_name, props, other):
        asyncio.create_task(self.__on_properties_changed__(props))

    def handle_change_playback_status(self, value: str):
        self.module_widget.set_playback_status(value)

    def handle_change_can_go_next(self, value: bool):
        if value:
            self.module_widget.set_on_next(self._on_next_callback)
        else:
            self.module_widget.set_on_next(None)

    def handle_change_can_go_previous(self, value: bool):
        if value:
            self.module_widget.set_on_previous(self._on_prev_callback)
        else:
            self.module_widget.set_on_previous(None)

    async def __on_properties_changed__(self, props: Dict[str, Any]):
        for key in props.keys():
            if key == "PlaybackStatus":
                playback_status = props[key].value
                assert isinstance(playback_status, str)
                self.handle_change_playback_status(playback_status)
            elif key == "CanGoNext":
                can_go_next = props[key].value
                assert isinstance(can_go_next, bool)
                self.handle_change_can_go_next(can_go_next)
            elif key == "CanGoPrevious":
                can_go_prev = props[key].value
                assert isinstance(can_go_prev, bool)
                self.handle_change_can_go_previous(can_go_prev)
            elif key == "Metadata":
                title = props[key].value.get("xesam:title")
                if title:
                    title = title.value
                self.modal_widget.set_title(title)

    def _on_select_player_callback(self, name):
        asyncio.create_task(self._select_player(name))

    async def _select_player(self, bus_name: Optional[str]):
        logging.info(f"binding with {bus_name}")
        obj_path = "/org/mpris/MediaPlayer2"
        mpris_interface_name = "org.mpris.MediaPlayer2.Player"
        props_interface_name = "org.freedesktop.DBus.Properties"

        if bus_name:
            introspection = await self.bus.introspect(bus_name, obj_path)
            proxy_object = self.bus.get_proxy_object(bus_name, obj_path, introspection)
            self.mpris_interface = proxy_object.get_interface(mpris_interface_name)
            self.mpris_props_interface = proxy_object.get_interface(
                props_interface_name
            )

            self.selected_player = bus_name
            self.mpris_props_interface.on_properties_changed(
                self._on_properties_changed_callback
            )
            self.module_widget.set_on_play_pause(self._on_play_pause_callback)
            self.modal_widget.set_title(await self._get_title())

            playback_status = await self.mpris_interface.get_playback_status()
            self.handle_change_playback_status(playback_status)

            can_go_next = await self.mpris_interface.get_can_go_next()
            self.handle_change_can_go_next(can_go_next)

            can_go_prev = await self.mpris_interface.get_can_go_previous()
            self.handle_change_can_go_previous(can_go_prev)
        else:
            self.mpris_interface = None
            self.mpris_props_interface = None
            self.selected_player = None
            self.module_widget.set_on_play_pause(None)
            self.handle_change_playback_status("Paused")
            self.handle_change_can_go_next(False)
            self.handle_change_can_go_prev(False)

    async def _get_title(self):
        metadata: Dict = await self.mpris_interface.get_metadata()
        title = metadata.get("xesam:title")
        if title:
            title = title.value
        return title

    async def __init_dbus__(self):
        bus_name = "org.freedesktop.DBus"
        obj_path = "/org/freedesktop/DBus"

        self.bus = await MessageBus(bus_type=BusType.SESSION).connect()
        introspection = await self.bus.introspect(bus_name, obj_path)
        proxy_object = self.bus.get_proxy_object(bus_name, obj_path, introspection)
        self.dbus_interface = proxy_object.get_interface("org.freedesktop.DBus")

        self.dbus_interface.on_name_owner_changed(self._on_name_owner_changed_callback)

        names = await self.dbus_interface.call_boxlist_names()
        for name in names:
            if name.startswith("org.mpris.MediaPlayer2"):
                self.bus_names.add(name)
        self.modal_widget.set_items(self.bus_names)


class MprisWidget(Gtk.Grid):
    def __init__(self, modal_menubutton, expander):
        super().__init__()
        # self.set_column_homogeneous(False)
        # self.set_row_homogeneous(False)

        self.button_play = Gtk.Button.new()
        self.button_play_image = Gtk.Image.new()
        self.button_play.set_image(self.button_play_image)
        self.button_play.set_relief(Gtk.ReliefStyle.NONE)
        self.button_play.set_sensitive(False)
        self.set_playback_status("Paused")
        Module.__remove_button_frame__(self.button_play)

        self.button_prev = Gtk.Button.new_from_icon_name(
            "media-skip-backward-symbolic", Gtk.IconSize.SMALL_TOOLBAR
        )
        self.button_prev.set_relief(Gtk.ReliefStyle.NONE)
        self.button_prev.set_sensitive(False)
        Module.__remove_button_frame__(self.button_prev)

        self.button_next = Gtk.Button.new_from_icon_name(
            "media-skip-forward-symbolic", Gtk.IconSize.SMALL_TOOLBAR
        )
        self.button_next.set_relief(Gtk.ReliefStyle.NONE)
        self.button_next.set_sensitive(False)
        Module.__remove_button_frame__(self.button_next)

        self.menubutton = modal_menubutton
        menubutton_image = Gtk.Image.new_from_icon_name(
            "go-down-symbolic", Gtk.IconSize.SMALL_TOOLBAR
        )
        self.menubutton.set_image(menubutton_image)
        Module.__remove_button_frame__(self.menubutton)
        self.menubutton.set_relief(Gtk.ReliefStyle.NONE)
        expander(self.menubutton)

        self.attach(self.button_prev, 0, 0, 1, 1)
        self.attach(self.button_play, 1, 0, 1, 1)
        self.attach(self.button_next, 2, 0, 1, 1)
        self.attach(self.menubutton, 0, 1, 3, 1)

    def set_playback_status(self, playback_status):
        if playback_status == "Playing":
            self.button_play_image.set_from_icon_name(
                "media-playback-pause-symbolic", Gtk.IconSize.SMALL_TOOLBAR
            )
        elif playback_status == "Paused":
            self.button_play_image.set_from_icon_name(
                "media-playback-start-symbolic", Gtk.IconSize.SMALL_TOOLBAR
            )

    def set_on_play_pause(self, on_play_pause: Optional[Callable]):
        if on_play_pause:
            self.on_play_pause = on_play_pause
            self.button_play_handler_id = self.button_play.connect(
                "clicked", lambda button: on_play_pause()
            )
            self.button_play.set_sensitive(True)
        else:
            self.on_play_pause = None
            self.button_play.disconnect(self.button_play_handler_id)
            self.button_play.set_sensitive(False)

    def set_on_next(self, on_next: Optional[Callable]):
        if on_next:
            self.on_next = on_next
            self.button_next_handler_id = self.button_next.connect(
                "clicked", lambda button: on_next()
            )
            self.button_next.set_sensitive(True)
        else:
            self.on_next = None
            if hasattr(self, "button_next_handler_id"):
                self.button_next.disconnect(self.button_next_handler_id)
            self.button_next.set_sensitive(False)

    def set_on_previous(self, on_prev: Optional[Callable]):
        if on_prev:
            self.on_prev = on_prev
            self.button_prev_handler_id = self.button_prev.connect(
                "clicked", lambda button: on_prev()
            )
            self.button_prev.set_sensitive(True)
        else:
            self.on_prev = None
            if hasattr(self, "button_prev_handler_id"):
                self.button_prev.disconnect(self.button_prev_handler_id)

            self.button_prev.set_sensitive(False)


class MprisModalWidget(Gtk.Box):
    def __init__(self, on_select):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.combo = Gtk.ComboBoxText()
        self.add(self.combo)
        self.on_select = on_select
        self.combo.connect("changed", self.on_select_changed)
        self.label_title = Gtk.Label()
        self.add(self.label_title)

    def on_select_changed(self, combo):
        active = self.combo.get_active_text()
        if active is not None:
            self.on_select(active)

    def set_title(self, title: Optional[str]):
        if title:
            self.label_title.set_label(f"Title: {title}")
        else:
            self.label_title.set_label("Title: Unknown")

    def set_items(self, items: Iterable):
        self.combo.remove_all()
        for item in items:
            self.combo.set_active_id(item)
            self.combo.append_text(item)



from gi.repository import Gtk
from cairo import Operator, FillRule
from python_reactive_ui import Children
from python_reactive_ui.backends.gtk3.builtin_component import Gtk3BuiltinComponent


class BatteryIcon(Gtk3BuiltinComponent):
    def _pre_init(self):
        self._set_widget(BatteryIconWidget())

    def _receive_props(self, new_props):
        if "charge" in new_props:
            self.gtk_widget.set_charge(new_props["charge"])
        if "ac" in new_props:
            self.gtk_widget.set_ac(new_props["ac"])
        if "size_request" in new_props:
            self.gtk_widget.set_size_request(*new_props["size_request"])

    def receive_children(self, new_children: Children):
        pass

    def _dismount(self):
        self._dismounter(self.gtk_widget)

    def _mount(self):
        self._mounter(self.gtk_widget)


class BatteryIconWidget(Gtk.DrawingArea):
    def __init__(self, charge: float = 0, ac: bool = False) -> None:
        super().__init__()
        self.charge = charge
        self.ac = ac
        self.connect("draw", lambda area, context: self.draw_battery(area, context))

    def set_charge(self, charge: float):
        self.charge = charge
        self.queue_draw()

    def set_ac(self, ac: bool):
        self.ac = ac
        self.queue_draw()

    def draw_battery(self, area, context):
        context.scale(area.get_allocated_width(), area.get_allocated_height())
        fg_color = area.get_style_context().get_color(Gtk.StateFlags.NORMAL)

        def draw_polygon(context, pts):
            x0, y0 = pts[0]
            context.move_to(x0, y0)
            for x, y in pts[1:]:
                context.line_to(x, y)
            context.close_path()

        battery_line_width = 0.05
        battery_width = 0.4
        battery_height = 0.60
        cap_height = 0.1
        cap_width = 0.2
        battery_x = 0.5 - battery_width / 2
        battery_y = 0.5 - battery_height / 2 + cap_height / 2
        cap_x = 0.5 - cap_width / 2

        pts_outer = [
            (battery_x, battery_y),
            (cap_x, battery_y),
            (cap_x, battery_y - cap_height),
            (cap_x + cap_width, battery_y - cap_height),
            (cap_x + cap_width, battery_y),
            (battery_x + battery_width, battery_y),
            (battery_x + battery_width, battery_y + battery_height),
            (battery_x, battery_y + battery_height),
        ]

        battery_inner_x = battery_x + battery_line_width
        battery_inner_y = battery_y + battery_line_width
        battery_inner_width = battery_width - 2 * battery_line_width
        battery_inner_height = (battery_height - 2 * battery_line_width) * (
            1 - self.charge
        )

        pts_inner = [
            (battery_inner_x, battery_inner_y),
            (battery_inner_x + battery_inner_width, battery_inner_y),
            (
                battery_inner_x + battery_inner_width,
                battery_inner_y + battery_inner_height,
            ),
            (battery_inner_x, battery_inner_y + battery_inner_height),
        ]

        context.set_line_width(0.05)
        if self.charge > 0.15 or self.ac:
            context.set_source_rgba(
                fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha
            )
        else:
            context.set_source_rgb(0.8, 0, 0)
        context.set_fill_rule(FillRule.EVEN_ODD)

        draw_polygon(context, pts_outer)
        context.clip()
        draw_polygon(context, pts_inner)
        context.new_sub_path()
        context.move_to(0, 0)
        context.line_to(1, 0)
        context.line_to(1, 1)
        context.line_to(0, 1)
        context.close_path()
        context.clip()
        context.paint()
        context.reset_clip()

        if self.ac:
            lightning_ratio = 0.5
            lightning_width = battery_width * lightning_ratio
            lightning_height = battery_height * lightning_ratio
            lightning_x = battery_x + (battery_width - lightning_width) / 2
            lightning_y = battery_y + (battery_height - lightning_height) / 2

            def adjust_point(p):
                x, y = p
                return (
                    lightning_width * x + lightning_x,
                    lightning_height * y + lightning_y,
                )

            pts = [
                adjust_point((0, 0.5)),
                adjust_point((0.75, 0)),
                adjust_point((0.625, 0.375)),
                adjust_point((1, 0.5)),
                adjust_point((0.25, 1)),
                adjust_point((0.375, 0.625)),
            ]
            context.set_line_width(0.03)
            draw_polygon(context, pts)
            context.set_source_rgba(
                fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha
            )
            context.set_operator(Operator.SOURCE)
            context.fill_preserve()
            context.set_source_rgba(
                1 - fg_color.red, 1 - fg_color.green, 1 - fg_color.blue, fg_color.alpha
            )
            context.stroke()



from gi.repository import Gtk


class Volume(Gtk.DrawingArea):
    def __init__(self, volume: float = 0) -> None:
        super().__init__()
        self.volume = volume
        self.connect("draw", lambda area, context: self.draw_battery(area, context))

    def update(self, volume: float):
        self.volume = volume
        self.queue_draw()

    def draw_battery(self, area, context):
        context.scale(area.get_allocated_width(), area.get_allocated_height())
        fg_color = area.get_style_context().get_color(Gtk.StateFlags.NORMAL)

        def draw_polygon(context, pts):
            x0, y0 = pts[0]
            context.move_to(x0, y0)
            for x, y in pts[1:]:
                context.line_to(x, y)
            context.close_path()

        volume = min([self.volume, 1])

        lw = 0.1
        margin = lw / 2

        outer_triangle = [
            (margin, 1 - margin),
            (1 - margin, margin),
            (1 - margin, 1 - margin),
        ]

        inner_triangle = [(0, 1), (volume, 1 - volume), (volume, 1)]

        context.set_line_width(0.1)
        context.set_source_rgba(
            fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha
        )
        draw_polygon(context, outer_triangle)
        context.stroke()
        draw_polygon(context, inner_triangle)
        context.fill()

        if self.volume > 1:
            v = self.volume % 1
            extra_triangle = [(0, 1), (v, 1 - v), (v, 1)]

            context.set_source_rgba(0.8, 0, 0, 1)
            draw_polygon(context, extra_triangle)
            context.fill()



import logging
from gi.repository import Notify
from typing import Optional


def notify_error(summary: Optional[str] = None, body: Optional[str] = None):
    if summary is None:
        summary = "ustatus error"
    if body is None:
        body = "An error has ocurred"

    logging.error(f"{summary}: {body}")

    notification = Notify.Notification.new(summary, body, "dialog-error")
    notification.set_urgency(2)
    notification.show()



from gi.repository import Gtk
from ustatus.module import Module
from ustatus.utils.power_utils import (
    shutdown,
    restart,
    suspend_disk,
    suspend_mem,
    suspend_freeze,
)


class PowerModule(Module):
    def __init__(
        self,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.modal_widget = PowerModalWidget()
        modal_menubutton = self.get_popover_menubutton(self.modal_widget)
        self.module_widget = PowerWidget(
            modal_menubutton, expander=lambda widget: self._expand_widthwise(widget)
        )

        self.set_module_widget(self.module_widget)

    def _update(self) -> bool:
        return True

    def _update_modal(self) -> bool:
        return True


class PowerWidget(Gtk.Box):
    def __init__(self, modal_menubutton, expander):
        super().__init__()

        expander(self)

        self.menubutton = modal_menubutton
        menubutton_image = Gtk.Image.new_from_icon_name(
            "system-shutdown-symbolic", Gtk.IconSize.SMALL_TOOLBAR
        )
        self.menubutton.set_image(menubutton_image)
        Module.__remove_button_frame__(self.menubutton)
        self.menubutton.set_relief(Gtk.ReliefStyle.NONE)
        expander(self.menubutton)

        self.set_center_widget(self.menubutton)


class PowerModalWidget(Gtk.Box):
    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=5)
        self.shutdown_button = Gtk.Button.new_with_label("Shutdown")
        self.shutdown_button.connect("clicked", lambda _: self.verify_action(shutdown))
        self.suspend_mem_button = Gtk.Button.new_with_label("Suspend (mem)")
        self.suspend_mem_button.connect(
            "clicked", lambda _: self.verify_action(suspend_mem)
        )
        self.suspend_freeze_button = Gtk.Button.new_with_label("Suspend (freeze)")
        self.suspend_freeze_button.connect(
            "clicked", lambda _: self.verify_action(suspend_freeze)
        )
        self.suspend_disk_button = Gtk.Button.new_with_label("Suspend (disk)")
        self.suspend_disk_button.connect(
            "clicked", lambda _: self.verify_action(suspend_disk)
        )
        self.restart_button = Gtk.Button.new_with_label("Restart")
        self.restart_button.connect("clicked", lambda _: self.verify_action(restart))

        self.add(self.suspend_freeze_button)
        self.add(self.suspend_mem_button)
        self.add(self.suspend_disk_button)
        self.add(self.restart_button)
        self.add(self.shutdown_button)

    def verify_action(self, action):
        dialog = ConfirmDialog(self.get_toplevel())
        response = dialog.run()
        dialog.destroy()

        if response == Gtk.ResponseType.YES:
            action()


class ConfirmDialog(Gtk.Dialog):
    def __init__(self, parent):
        super().__init__(title="Confirmation", transient_for=parent, flags=0)
        self.set_resizable(False)
        self.add_buttons(
            Gtk.STOCK_NO, Gtk.ResponseType.NO, Gtk.STOCK_YES, Gtk.ResponseType.YES
        )
        self.set_default_size(150, 100)
        label = Gtk.Label(label=f"Are you sure?")
        box = self.get_content_area()
        box.add(label)
        self.show_all()



from time import localtime, strftime
from typing import Callable
from python_reactive_ui import Component, use_state
from python_reactive_ui.backends.gtk3.builtin.label import Label
from ustatus.hooks.use_timer import use_timer


class Clock(Component):
    def _update_time(self, set_time: Callable):
        set_time(localtime())

    def _render(self, props, children):
        time, set_time = use_state(self, localtime())
        handle = use_timer(self, 1, lambda: self._update_time(set_time))
        return Label({"text": strftime("%H:%M:%S", time), "hexpand": True})



from collections import namedtuple
from gi.repository import Gtk
from python_reactive_ui import Component, use_state
from python_reactive_ui.backends.gtk3.builtin.box import Box
from python_reactive_ui.backends.gtk3.builtin.label import Label
from ustatus.components.drawing.battery_icon import BatteryIcon
from ustatus.hooks.use_timer import use_timer
import psutil


class Battery(Component):
    BatteryData = namedtuple("BatteryData", ["charge", "ac"])

    @classmethod
    def _get_data(cls):
        battery_data = psutil.sensors_battery()
        return (
            cls.BatteryData(
                charge=battery_data.percent / 100, ac=battery_data.power_plugged
            )
            if battery_data
            else cls.BatteryData(charge=0, ac=True)
        )

    def _render(self, props, children):
        data, set_data = use_state(self, Battery._get_data())
        use_timer(
            self,
            timer_secs=5,
            repeat=True,
            callback=lambda: set_data(Battery._get_data()),
        )
        return Box(
            { "halign": "center", "hexpand": True },
            [
                BatteryIcon(
                    {"charge": data.charge, "ac": data.ac, "size_request": (25, 25)}
                ),
                Label({"text": f"{data.charge * 100:.0f}%"}),
            ],
        )



from functools import reduce
from typing import Dict, Iterable
from ustatus.schema import bar, cmdline_friendly, module


def generate_docs():
    docs = f"""
# Configuration guide
Configuration is written in [TOML](https://toml.io/en/). Ustatus will read a file called `ustatus.toml` at `$XDG_CONFIG_HOME/ustatus/ustatus.toml` or at `$HOME/.config/ustatus/ustatus.toml` if the former is not defined.

## Bar configuration
All *bar configurations* should be under a table called `bars`. For example, a bar called `mystatusbar` is configured as in the following snippet:
```toml
[bars.mystatusbar]
modules_center = [
  "cpu",
  "tray",
  "power",
]
anchors = ["right"]
orientation = "vertical"
separators = true
width = 90
```
As long as the included modules are also defined (see [module configuration section](#module-configuration)), the bar can be started using
```bash
ustatus mystatusbar
```
Any bar configuration setting can be overriden with commandline flags; run
```bash
ustatus --help
```
for more details.

The available bar configuration keys are:

{generate_table(bar)}

## Module configuration
All *module configurations* should be under a table called `modules`. For example, a battery meter module called `mymodule` is configured as in the following snippet:
```toml
[modules.mymodule]
type = "battery"
show_label = true
label = "My Battery Module"
```

The available module configuration options are:

{generate_table(module)}
"""
    with open("CONFIGURATION.md", "w") as file:
        file.write(docs)


def generate_table(table):
    bar_table = " Key | Type | Available in command line | Default | Description \n ---|---|---|---|---\n"
    for prop_name, prop_details in table["properties"].items():
        bar_table += f"`{prop_name}` | {generate_type_doc(prop_details)} | {'yes' if cmdline_friendly(prop_details) else 'no'} | `{prop_details.get('default', None)}` | {prop_details.get('description', None)}\n"
    return bar_table


def generate_type_doc(type: Dict) -> str:
    if "enum" in type:
        return _intercalate(map(lambda a: f'`"{a}"`', type["enum"]), " *or* ")
    elif "type" in type:
        return type["type"]
    elif "anyOf" in type:
        return _intercalate(
            map(lambda a: generate_type_doc(a), type["anyOf"]),
            " *or* ",
        )
    else:
        return "unknown"


def _intercalate(l: Iterable[str], sep: str):
    out = ""
    for i, val in enumerate(l):
        if i != 0:
            out += sep
        out += val
    return out



import ustatus
from ustatus.module import Module
from dbus_next.service import ServiceInterface, method, dbus_property, signal
from dbus_next.signature import Variant
from dbus_next.aio.message_bus import MessageBus
import asyncio
import logging


async def init_service(on_hide, on_show, bar_name):
    bus = await MessageBus().connect()
    interface = UstatusRemoteService("ustatus.UstatusRemoteService", on_show, on_hide)
    bus.export("/UstatusRemoteService", interface)
    # now that we are ready to handle requests, we can request name from D-Bus
    asyncio.create_task(bus.request_name(f"ustatus.UstatusRemoteService.{bar_name}"))
    logging.info("Remote service initialized")


class UstatusRemoteService(ServiceInterface):
    def __init__(self, name, on_show, on_hide):
        super().__init__(name)
        self.on_show = on_show
        self.on_hide = on_hide
        self.is_shown = True

    @method()
    def Hide(self):
        logging.info("Hide called")
        self.on_hide()
        self.is_shown = False

    @method()
    def Show(self):
        logging.info("Show called")
        self.on_show()
        self.is_shown = True

    @method()
    def ToggleVisible(self):
        logging.info("Toggle called")
        if self.is_shown:
            self.on_hide()
        else:
            self.on_show()
        self.is_shown = not self.is_shown



import asyncio
import logging
from ustatus.config import ModuleConfig
from ustatus.module import Module
from typing import Callable, Dict, Iterable, Optional
from gi.repository import Gtk

from ustatus.utils.swaymsg import get_workspaces, subscribe, sway_command


class SwayModule(Module):
    def __init__(
        self,
        output: Optional[str],
        **kwargs,
    ):
        super().__init__(**kwargs)
        module_widget = SwayModuleWidget(
            gtk_orientation=self.gtk_orientation, on_select=self.on_select
        )
        self.set_module_widget(module_widget=module_widget)
        self.output = output
        asyncio.create_task(self._init_async())

    def on_select(self, num):
        asyncio.create_task(sway_command(f"workspace number {num}"))

    async def _init_async(self):
        workspaces = await self._get_workspaces()

        self.module_widget.setup_workspaces(workspaces)

        async for event in subscribe(["workspace"]):
            match event["change"]:
                case "focus":
                    if self._check_output(event["current"]):
                        self.module_widget.focus_workspace(event["current"])
                case "rename":
                    if self._check_output(event["current"]):
                        self.module_widget.rename_workspace(event["current"])
                case "init":
                    if self._check_output(event["current"]):
                        self.module_widget.init_workspace(event["current"])
                case "empty":
                    if self._check_output(event["current"]):
                        self.module_widget.remove_workspace(event["current"])
                case "move":
                    if self._check_output(event["current"]):
                        self.module_widget.move_workspace(event["current"])
                case "reload":
                    self.module_widget.setup_workspaces(await self._get_workspaces())

    def _check_output(self, workspace):
        return (not self.output) or workspace["output"] == self.output

    async def _get_workspaces(self) -> Iterable[Dict]:
        workspaces = await get_workspaces()
        if self.output:
            workspaces = filter(lambda w: w["output"] == self.output, workspaces)
        return workspaces


class WorkspaceLabel(Gtk.Label):
    def __init__(self, workspace):
        super().__init__(label=workspace["name"])
        self.workspace = workspace

    def get_num(self):
        return self.workspace["num"]

    def get_id(self):
        return self.workspace["id"]


class SwayModuleWidget(Gtk.FlowBox):
    def __init__(
        self,
        gtk_orientation: Gtk.Orientation,
        on_select: Callable,
        initial_workspaces: Optional[Iterable[Dict]] = None,
    ):
        match gtk_orientation:
            case Gtk.Orientation.HORIZONTAL:
                orientation = Gtk.Orientation.VERTICAL
            case Gtk.Orientation.VERTICAL:
                orientation = Gtk.Orientation.HORIZONTAL
            case other:
                raise Exception(f"Orientation {other} not recognized")
        super().__init__(orientation=orientation)
        self.on_select = on_select
        self.workspaces: Dict = dict()
        self.set_sort_func(
            lambda child1, child2, *user_data: -1
            if child1.get_child().get_num() < child2.get_child().get_num()
            else 1
        )
        self.set_homogeneous(True)
        self.connect("child-activated", self._on_child_activated)

        if initial_workspaces:
            self.setup_workspaces(initial_workspaces)

    def setup_workspaces(self, initial_workspaces):
        self._clear()
        for ws in initial_workspaces:
            self.init_workspace(ws)
            if ws["focused"]:
                self.focus_workspace(ws)
        self.show_all()

    def init_workspace(self, workspace):
        id = workspace["id"]
        if id in self.workspaces:
            raise Exception(
                f"Workspace {workspace['id']}:{workspace['num']}:{workspace['name']} already exists"
            )
        label = WorkspaceLabel(workspace=workspace)
        label.show()
        self.workspaces[id] = label
        self.add(label)

    def move_workspace(self, workspace):
        """Workspace is moved into current output"""
        id = workspace["id"]
        if id not in self.workspaces:
            label = WorkspaceLabel(workspace=workspace)
            label.show()
            self.workspaces[id] = label
            self.add(label)

    def remove_workspace(self, workspace):
        if workspace["id"] not in self.workspaces:
            raise Exception(
                f"Workspace {workspace['id']}:{workspace['num']}:{workspace['name']} does not exist, can't be removed!"
            )
        label = self.workspaces.pop(workspace["id"])
        self.remove(label)

    def rename_workspace(self, workspace):
        id = workspace["id"]
        if id not in self.workspaces:
            raise Exception(
                f"Workspace {workspace['id']}:{workspace['num']}:{workspace['name']} does not exist, can't be renamed!"
            )
        self.workspaces[id].set_label(workspace["name"])

    def focus_workspace(self, workspace):
        id = workspace["id"]
        if id not in self.workspaces:
            raise Exception(
                f"Workspace {workspace['id']}:{workspace['num']}:{workspace['name']} does not exist, can't be focused!"
            )
        self.select_child(self._find_child_with_label(self.workspaces[id]))

    def _find_child_with_label(self, label):
        for child in self.get_children():
            if child.get_child() == label:
                return child
        raise Exception("Child not found")

    def _clear(self):
        while True:
            try:
                _, label = self.workspaces.popitem()
                self.remove(label)
            except KeyError:
                break

    def _on_child_activated(self, flow_box, child):
        self.on_select(child.get_child().get_num())



import asyncio, gbulb, logging, logging.handlers
from multiprocessing import Process
from python_dbus_system_api import start_server

from ustatus.ustatus import Ustatus

from ustatus.utils.notifications import notify_error


def main():
    setup_logging()
    # Ensure API is running (if already running this will just queue to get the
    # bus name)
    api_process = Process(target=start_server)
    api_process.start()
    gbulb.install(gtk=True)  # only necessary if you're using GtkApplication
    application = Ustatus()
    loop = asyncio.get_event_loop()
    try:
        loop.run_forever(application=application)
    except Exception as e:
        notify_error(summary="Uncaught error", body=f"e")


def setup_logging():
    logging.basicConfig(
        level=logging.INFO,
        style="{",
        format="[{levelname}({name}):{filename}:{funcName}] {message}",
    )
    root_logger = logging.getLogger()
    sys_handler = logging.handlers.SysLogHandler(address="/dev/log")
    root_logger.addHandler(sys_handler)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ustatus',
 'ustatus.components',
 'ustatus.components.drawing',
 'ustatus.components.modules',
 'ustatus.graphics',
 'ustatus.hooks',
 'ustatus.modules',
 'ustatus.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyGObject>=3.42.0,<4.0.0',
 'dbus-next>=0.2.3,<0.3.0',
 'gbulb>=0.6.2,<0.7.0',
 'jsonschema>=4.4.0,<5.0.0',
 'psutil==5.8.0',
 'pulsectl-asyncio>=0.2.0,<0.3.0',
 'pulsectl>=22.3.2,<23.0.0',
 'python-dbus-system-api>=0.1.0,<0.2.0',
 'python-reactive-ui>=0.2.9,<0.3.0',
 'tomli>=1.2.2,<2.0.0']

entry_points = \
{'console_scripts': ['ustatus = ustatus.main:main',
                     'ustatus_docgen = ustatus.docgen:generate_docs']}

setup_kwargs = {
    'name': 'ustatus',
    'version': '0.1.10',
    'description': 'GTK-based status window for wayland shells.',
    'long_description': "# ustatus\n\n_NOTE: Previously named pystatus, renamed due to a PyPI project naming conflict\npreventing me from publishing the project._\n\nConfigurable status window and/or bar for wayland compositors with support for\nthe wlr-layer-shell protocol.\n\n![Screenshot closed](pystatus_closed.png)\n![Screenshot cpu meters](pystatus_cpu.png)\n![Screenshot power-profiles-daemon](pystatus_ppd.png)\n![Screenshot powermenu](pystatus_power.png)\n\n## Features\n\n- TOML configuration.\n  - Supports configuration of different bars, and choice of bar at startup with\n    commandline argument.\n- A variety of built-in modules:\n  - System tray module ([StatusNotifierItem protocol](https://www.freedesktop.org/wiki/Specifications/StatusNotifierItem/)), with Dbusmenu.\n  - Cpu monitor module, with usage graphs.\n  - MPRIS module.\n  - Battery module.\n  - Volume module.\n  - [power-profiles-daemon](https://gitlab.freedesktop.org/hadess/power-profiles-daemon) remote module.\n- Remote interface through dbus, currently allowing:\n  - Show/hide/toggle bar or status window.\n\n## Running without installing\n\nThe project is built using [Poetry](https://python-poetry.org/).\nTo run, clone the project\n\n```\ngit clone https://github.com/mbrea-c/ustatus.git\n```\n\nmove to the cloned directory\n\n```\ncd ustatus\n```\n\ninstall dependencies in a local venv with\n\n```\npoetry install\n```\n\nand run ustatus with\n\n```\npoetry run ustatus <bar name>\n```\n\n## User-wide installation from source\n\nThe project is built using [Poetry](https://python-poetry.org/).\n\nTo build\nlocally, clone the project\n\n```\ngit clone https://github.com/mbrea-c/ustatus.git\n```\n\nmove to the cloned directory\n\n```\ncd ustatus\n```\n\nand build\n\n```\npoetry build\n```\n\nThis will create two files in the `dist` directory,\n\n```\ndist/ustatus-<version_number>.tar.gz\ndist/ustatus-<version_number>-py3-none-any.whl\n```\n\nFor a user-wide installation, run\n\n```\ncd dist\npip install --upgrade ustatus-<version_number>.tar.gz\n```\n\n## Configuration\n\nSee the [configuration guide](CONFIGURATION.md) for details. An example configuration file can be\nfound in [examples/ustatus.toml](examples/ustatus.toml)\n\n## Is this in an usable state?\n\nI think so, I have used (and still use) this in my daily-driver machine for about 6 months.\nHowever, there are currently some limitations that some might consider deal\nbreaking:\n\n- The UI respects the system GTK theme, but it is not possible to do separate\n  custom theming at the moment. I'm thinking of adding some form of this soon though, as GTK\n  allows theming using CSS fairly easily.\n- In my daily driver setup I use a status window anchored to the right side of\n  the screen, so I haven't done a lot of testing with a horizontal status _bar_.\n  It is possible, but there might (will) be bugs and rough edges. I'm willing to\n  work on this if there is interest.\n",
    'author': 'Manuel Brea',
    'author_email': 'm.brea.carreras@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mbrea-c/ustatus',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import subprocess


def shutdown():
    subprocess.run(["doas", "shutdown", "-h", "now"])


def restart():
    subprocess.run(["doas", "shutdown", "-r", "now"])


def suspend_mem():
    subprocess.run(["doas", "zzz", "-z"])


def suspend_freeze():
    subprocess.run(["doas", "zzz", "-S"])


def suspend_disk():
    subprocess.run(["doas", "zzz", "-Z"])



from python_reactive_ui import Component
from python_reactive_ui.backends.gtk3.builtin.label import Label


class Test(Component):
    def _render(self, props, children):
        return Label({"text": "Testing..."})



from typing import Callable, Optional
from gi.repository import Gtk, GLib
from python_reactive_ui import Component
from python_reactive_ui.backends.gtk3.builtin.box import Box
from python_reactive_ui.backends.gtk3.builtin.label import Label

from ustatus.config import ModuleConfig


class ReactiveModule(Component):
    def _render(self, props, children):
        module_config = props["module_config"]
        module_type = module_config.type
        match children:
            case [child]:
                return Box(
                    {"css_classes": ["module", module_type], "orientation": "vertical"},
                    [
                        *(
                            (Label({"text": module_config.label}),)
                            if module_config.show_label
                            else ()
                        ),
                        child,
                    ],
                )
            case _:
                return ValueError("Only a single child allowed")



import asyncio
import json


async def run(cmd):
    proc = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    match proc.returncode:
        case 0:
            pass
        case 1:
            raise Exception(f"swaymsg exception: {stderr.decode()}")
        case 2:
            raise Exception(f"sway exception: {stderr.decode()}")

    return json.loads(stdout.decode())


async def run_stream(cmd):
    proc = await asyncio.create_subprocess_exec(*cmd, stdout=asyncio.subprocess.PIPE)

    assert proc.stdout is not None

    buf = bytes()
    open_obj = 0
    open_lst = 0
    while True:
        # TODO: Make or find a more efficient buffer for this purpose
        c = await proc.stdout.read(n=1)
        if not c:
            break
        match c:
            case b"{":
                open_obj += 1
            case b"}":
                open_obj -= 1
            case b"[":
                open_lst += 1
            case b"]":
                open_lst -= 1
        buf += c
        if open_obj == 0 and open_lst == 0 and not c.isspace():
            yield json.loads(buf)
            buf = bytes()


async def get_workspaces():
    return await run(["swaymsg", "-t", "get_workspaces"])


async def get_outputs():
    return await run(["swaymsg", "-t", "get_outputs"])


async def sway_command(cmd: str):
    return await run(["swaymsg", cmd])


async def subscribe(event_types):
    async for event in run_stream(
        ["swaymsg", "--monitor", "-t", "subscribe", json.dumps(event_types)]
    ):
        yield event



from typing import Callable
from dbus_next.aio.message_bus import MessageBus
from gi.repository import Gtk, GLib
from pulsectl import pulsectl
import pulsectl_asyncio, asyncio
from python_dbus_system_api.interfaces.volume import VolumeInterface

from python_reactive_ui import Component, use_state
from python_reactive_ui.backends.gtk3.builtin.box import Box
from python_reactive_ui.backends.gtk3.builtin.label import Label
from python_reactive_ui.backends.gtk3.builtin.progress_bar import ProgressBar
from python_reactive_ui.lib.hooks import use_effect
from ustatus.config import ModuleConfig
from ustatus.graphics.meter import Meter
from ustatus.graphics.volume import Volume
from ustatus.module import Module


class Volume(Component):
    async def _setup_client(self, set_client, set_sinks):
        name = "pysysapi.api"
        path = "/pysysapi"
        bus = await MessageBus().connect()
        introspection = await bus.introspect(name, path)
        proxy_object = bus.get_proxy_object(name, path, introspection)
        interface = proxy_object.get_interface(VolumeInterface.INAME)
        interface.on_event(
            lambda event: asyncio.create_task(self._update_sinks(interface, set_sinks))
        )
        await self._update_sinks(interface, set_sinks)
        set_client(interface)

    async def _update_sinks(self, client, set_sinks):
        set_sinks(await client.call_get_sinks())

    def _render(self, props, children):
        sinks, set_sinks = use_state(self, dict())
        client, set_client = use_state(self, None)
        use_effect(
            self,
            lambda: asyncio.create_task(self._setup_client(set_client, set_sinks)),
            [],
        )

        return Box(
            {"hexpand": True, "halign": "fill", "size_request": (-1, 60)},
            [
                Box(
                    {
                        "css_classes": ["sink"],
                        "orientation": "vertical",
                        "opacity": 0.5 if sink["mute"].value == 1 else 1.0,
                        "tooltip_text": sink["description"].value,
                    },
                    [
                        ProgressBar(
                            {
                                "orientation": "vertical",
                                "fraction": sink["volume_all_chans"].value,
                                "vexpand": True,
                                "inverted": True,
                                "hexpand": True,
                                "halign": "center",
                            }
                        ),
                        Label({"text": f"{100 * sink['volume_all_chans'].value:.0f}%"}),
                    ],
                )
                for sink in sinks
            ],
        )



from typing import Optional
from gi.repository import Gtk, Gdk
from collections import deque


class LineGraph(Gtk.DrawingArea):
    def __init__(
        self,
        n_values=100,
        color: Optional[Gdk.RGBA] = None,
        min: float = 0,
        max: float = 1,
    ) -> None:
        super().__init__()
        self.min = min
        self.max = max
        self._update_range()
        self.n_values = n_values
        self.values = deque([0] * n_values, maxlen=n_values)
        self.color = color
        self.connect("draw", lambda area, context: self.draw_line_graph(area, context))

    def set_max(self, max):
        self.max = max
        self._update_range()

    def set_min(self, min):
        self.min = min
        self._update_range()

    def _update_range(self):
        self.range = self.max - self.min

    def set_values(self, new_values):
        self.values = deque(new_values, maxlen=self.n_values)
        self.queue_draw()

    def push_value(self, new_value):
        self.values.append(new_value)
        self.queue_draw()

    def draw_line_graph(self, area, context):
        context.scale(area.get_allocated_width(), area.get_allocated_height())
        if not self.color:
            fg_color = area.get_style_context().get_color(Gtk.StateFlags.NORMAL)
        else:
            fg_color = self.color
        context.set_source_rgba(
            fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha
        )

        maxsize = max([area.get_allocated_width(), area.get_allocated_height()])
        context.set_line_width(1 / maxsize)

        separation: float = 1 / (len(self.values) - 1)

        for i, val in enumerate(self.values):
            x = i * separation
            height = (val - self.min) / self.range

            context.line_to(x, 1 - height)

        context.stroke_preserve()

        context.line_to(1, 1)
        context.line_to(0, 1)

        context.set_source_rgba(
            fg_color.red, fg_color.green, fg_color.blue, fg_color.alpha / 2
        )
        context.fill()
